package org.jboss.seam.selectitems.annotations;

import static java.lang.annotation.ElementType.FIELD;
import static java.lang.annotation.ElementType.METHOD;
import static java.lang.annotation.RetentionPolicy.RUNTIME;

import java.lang.annotation.Documented;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;

import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.DataBinderClass;
import org.jboss.seam.selectitems.databinding.SelectItemsBinder;

/**
 * Outjects a List to the same scope as the owning component
 * (or to the EVENT scope in the case of a stateless component),
 * after converting it to a list of SelectItems. Note that the List
 * will be re-wrapped and re-outjected each time the current
 * component value is different to the value held by the
 * context variable as determined by calling List.equals().
 * 
 * N.B. A getter method 'getSeam' has a name 'seam'
 * 
 * @author Peter Muir
 * @author <a href="mailto:jh2000guard-jboss@yahoo.com">Jim Hazen</a>
 * 
 */
@Target({FIELD, METHOD})
@Retention(RUNTIME)
@Documented
@DataBinderClass(SelectItemsBinder.class)
public @interface SelectItems
{
	public enum Strategy
    {
        /**
         * The INDEX strategy indicates that the rendered JSF value
         * should be the index of the original list item.
         */
        INDEX,
        
        /**
         * The OBJECT strategy indicates that the rendered JSF value
         * should be the actual original item object.  This requires a
         * custom converter; the Entity converter can be used if the list
         * is loaded from the persistence context.
         * */
        OBJECT,
        
        /**
         * The STRING strategy indicates that the rendered JSF value should
         * be a String.
         * @link org.jboss.seam.databinding.selectitems.SelectItemsStringStrategy
         */
        STRING;
    }
	
   /**
    * The context variable name. Defaults to the name of 
    * the annotated field or getter method.
    */
   String value() default "";
   
   /**
    * By default, the SelectItems list is outjected to the same
    * scope as the owning component, or to the EVENT scope
    * in the case of a stateless component. If scope=ScopeType.PAGE
    * is explicitly specified, the DataModel will be kept
    * in the PAGE context instead.
    * 
    * @return ScopeType.UNSPECIFIED or ScopeType.PAGE
    */
   ScopeType scope() default ScopeType.UNSPECIFIED;
   
   /**
    * The name of the getter method to use for the label of the SelectItem.
    * Defaults to the toString() value.
    * The label is looked up in the seam message bundle as a key, if a value
    * is found that will be used, otherwise the key is used
    */
   String label() default "";
   
   /**
    * The name of the getter method to set the disabled property of the SelectItem.
    * By default the SelectItem is enabled.
    */
   String disabled() default "";
   
   /**
    * Choose a strategy for automatic value conversion.
    * Defaults to OBJECT.
    * A suitable converter needs to be defined unless the object is a string
    */
   Strategy strategy() default Strategy.OBJECT;
   
   /**
    * The name of the getter method to use for the JSF value. 
    * Ignored usless strategy==Strategy.STRING, defaults to toString();
    */
   String jsfValue() default "";
   
   /**
    * Adds a "none"-selection to the top of the selection list if set. 
    * 
    * The string value is looked up in the seam message bundle, if a value
    * is found that will be used, otherwise the key is used
    */
   String addNoSelectionLabel() default "";
   
   /**
    * Disables/enables use of the Seam message bundle when interpreting labels.
    * By default true; the label is used as a key for looking up a string to display.
    * If there is no entry for the key, the key is displayed
    */
   boolean useMessageBundle() default true;
}