package org.jboss.seam.selectitems.jsf;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import javax.faces.application.FacesMessage;
import javax.faces.component.StateHolder;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.convert.Converter;
import javax.faces.convert.ConverterException;
import javax.faces.el.ValueBinding;
import javax.persistence.Entity;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Id;

import org.apache.commons.beanutils.ConvertUtils;
import org.jboss.seam.core.FacesMessages;
import org.jboss.seam.selectitems.SelectItemsConfig;
import org.jboss.seam.util.Naming;

/**
 * Allows generic conversion of EJB3 Entities.
 * 
 * By default the the class of the entity is inferred from the value attribute of the parent
 * component.  It can be overriden by the optional entityClass attribute.
 * 
 * See README for instructions and ideas on how to register the converter
 * 
 * @author Peter
 *
 */
public class EntityConverter implements Converter, StateHolder {

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.faces.convert.Converter#getAsString(javax.faces.context.FacesContext,
	 *      javax.faces.component.UIComponent, java.lang.Object)
	 */
	public String getAsString(FacesContext faces, UIComponent c, Object obj) {
		if (obj == null)
			return SelectItemsConfig.NO_SELECTION_VALUE;
		// For some reason MyFaces will sometimes try to reconvert an already converted object so
		// pass through
		if (obj instanceof String) 
			return (String) obj;
		return getEntityIdAsString(obj);
	}
	
	/**
	 * Utility method to get the ID of the entity as a string
	 * @param entity The entity to use
	 * @return The ID of the entity as a string
	 */
	private String getEntityIdAsString(Object entity) {
		Field idField = findIdField(entity);
		Method idMethod = findIdProperty(entity);
		if (idField != null) {
			try {
				return idField.get(entity).toString();
			} catch (IllegalArgumentException e) {
				throw new ConverterException(
						getMessage("Unable to get id from entity"), e);
			} catch (IllegalAccessException e) {
				throw new ConverterException(
						getMessage("Unable to get id from entity"), e);
			}
		} else if (idMethod != null) {
			Object[] params = {};
			try {
				return idMethod.invoke(entity, params).toString();
			} catch (IllegalArgumentException e) {
				throw new ConverterException(
						getMessage("Unable to get id from entity"), e);
			} catch (IllegalAccessException e) {
				throw new ConverterException(
						getMessage("Unable to get id from entity"), e);
			} catch (InvocationTargetException e) {
				throw new ConverterException(
						getMessage("Unable to get id from entity"), e);
			}
		} else {
			throw new ConverterException(
					getMessage("Unable to get id from entity"));
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.faces.convert.Converter#getAsObject(javax.faces.context.FacesContext,
	 *      javax.faces.component.UIComponent, java.lang.String)
	 */
	public Object getAsObject(FacesContext facesContext, UIComponent c, String value) {
		if (SelectItemsConfig.NO_SELECTION_VALUE.equals(value)) return null;
		Class clazz = getComponentClass(c, facesContext);
		return getEntity(clazz, getEntityIdAsObject(value, clazz));
	}


	/**
	 * Utility method to determine class of the entity to load
	 * If class parameter has use that otherwise try to infer from the 'value' attribute
	 * @param cmp The component to which the converter is attached
	 * @param facesContext
	 * @return The class of the entity to load
	 */
	private Class getComponentClass(UIComponent cmp, FacesContext facesContext) {
		// If the class parameter has been specified use it
		if (entityClazz != null && !entityClazz.equals("") ) {
			try {
				return Class.forName(entityClazz);
			} catch (ClassNotFoundException e) {
				throw new ConverterException(getMessage(entityClazz + " cannot be loaded").getDetail(), e);
			}
		} else {
			// Otherwise try to find out by value binding
			ValueBinding vb = cmp.getValueBinding("value");
			if (vb == null) {
				throw new ConverterException(
						getMessage("Unable to determine entity from value attribute"));
			}
			
			return vb.getType(facesContext);
		}
	}
	
	
	
	
	/**
	 * Utility method to retrieve the Entity from the database
	 * @param clazz The class of the entity to load
	 * @param id The id of the entity to load
	 * @return The entity
	 */
	@SuppressWarnings("unchecked")
	private Object getEntity(Class clazz, Object id) {
		try {		
			EntityManagerFactory emf = (EntityManagerFactory) Naming.getInitialContext().lookup(getPersistenceUnitJndiName());
			EntityManager em = emf.createEntityManager();
			Object entity = em.find(clazz, id);
			if (entity == null) {
				throw new ConverterException(
						getMessage("Invalid selection, please try again"));
			} else {
				return entity;
			}
		} catch (Exception e) {
			throw new ConverterException(
					getMessage("Unable to access entity manager"), e);
		}
	}

	/**
	 * Utility method.  Given an entity and its id as a string return its Id as Object of the
	 * correct type
	 * @param id The Id of the entity
	 * @param clazz The class of the entity
	 * @return The id of the entity as a correctly typed object
	 */
	private Object getEntityIdAsObject(String id, Class clazz) {
		Field idField = findIdField(clazz);
		Method idMethod = findIdMethod(clazz);
		Class idClass;
		if (idField != null) {
			idClass = idField.getType();
		} else if (idMethod != null) {
			idClass = idMethod.getReturnType();
		} else {
			throw new ConverterException(
					getMessage("Unable to read id from entity"));
		}
		return ConvertUtils.convert(id, idClass);
	}

	/**
	 * Utility to find Id field on an entity
	 * @param entity The entity to find the id field on
	 * @return The Id field
	 */
	private Field findIdField(Object entity) {
		Class clazz = entity.getClass();
		return findIdField(clazz);
	}

	/**
	 * Utility to find Id field on an entity
	 * @param clazz The clazz of the entity to find the id field on
	 * @return The Id field
	 */
	@SuppressWarnings("unchecked")
	private Field findIdField(Class clazz) {
		// Check that the class passed is an entity
		if (!(clazz.isAnnotationPresent(Entity.class))) {
			throw new ConverterException(getMessage(clazz.getName() + " is not an entity"));
		}

		for (Class currClazz = clazz; !currClazz.equals(Object.class); currClazz = currClazz.getSuperclass()) {
			// Iterate over the fields and find the Id field
			for (Field f : currClazz.getDeclaredFields()) {
				if (!(f.isAccessible()))
					f.setAccessible(true);
				if (f.isAnnotationPresent(Id.class)) {
					return f;
				}
			}
		}
		return null;
	}
	
	/**
	 * Utility to find Id mproperty on an entity
	 * @param entity The entity to find the id property
	 * @return The Id property
	 */
	private Method findIdProperty(Object entity) {
		return findIdMethod(entity.getClass());
	}

	/**
	 * Utility to find Id property on an entity
	 * @param clazz The entity to find the id property on
	 * @return The Id property
	 */
	@SuppressWarnings("unchecked")
	private Method findIdMethod(Class clazz) {
		// Check that the class passed is an entity
		if (!(clazz.isAnnotationPresent(Entity.class))) {
			throw new ConverterException(getMessage(clazz.getName() + " is not an entity"));
		}

		for (Class currClazz = clazz; !currClazz.equals(Object.class); currClazz = currClazz.getSuperclass()) { 
			// Iterate over the fields and find the Id field
			for (Method m : currClazz.getDeclaredMethods()) {
				if (!(m.isAccessible()))
					m.setAccessible(true);
				if (m.isAnnotationPresent(Id.class)) {
					return m;
				}
			}
		}
		return null;
	}

	/**
	 * Utitliy method to get a FacesMessage for ConverterExceptions
	 * @param message The message to create
	 * @return The FacesMessage Object
	 */
	private FacesMessage getMessage(String message) {
		return FacesMessages.createFacesMessage(FacesMessage.SEVERITY_FATAL,
				message);
	}

	/*
	 * State saving implementation allowing parameters to be passed
	 *  (non-Javadoc)
	 * @see javax.faces.component.StateHolder#saveState(javax.faces.context.FacesContext)
	 */
	
	private String entityClazz;
	private String persistenceUnitJndiName;

	private boolean _transient;
	
	public Object saveState(FacesContext arg0) {
		Object[] values = new Object[1];
		values[0] = entityClazz;
		return values;
	}

	public void restoreState(FacesContext arg0, Object state) {
		Object[] values = (Object[])state;
		entityClazz = (String) values[0];
	}

	public boolean isTransient() {
		return _transient;
	}

	public void setTransient(boolean _transient) {
		this._transient = _transient;
	}

	public String getEntityClass() {
		return entityClazz;
	}

	public void setEntityClass(String clazz) {
		this.entityClazz = clazz;
	}

	public String getPersistenceUnitJndiName() {
		if (persistenceUnitJndiName == null) {
			return SelectItemsConfig.instance().getPersistenceUnitJndiName();
		} else {
			return persistenceUnitJndiName;
		}
	}

	public void setPersistenceUnitJndiName(String emfJndiName) {
		this.persistenceUnitJndiName = emfJndiName;
	}

}
